@@ -4,13 +4,62 @@ from typing import Dict, List, Optional
 from kh_common.auth import KhUser
 from kh_common.caching import AerospikeCache
 from kh_common.caching.key_value_store import KeyValueStore
-from kh_common.exceptions.http_error import BadRequest, NotFound
-from kh_common.scoring import confidence
-from kh_common.scoring import controversial as calc_cont
-from kh_common.scoring import hot as calc_hot
+from kh_common.exceptions.http_error import BadRequest
 from kh_common.sql import SqlInterface
 
 from fuzzly_posts.models import PostId, Score
+from math import log10, sqrt
+from typing import Union
+
+from scipy.stats import norm
+
+from kh_common.config.constants import epoch
+
+
+"""
+resources:
+	https://github.com/reddit-archive/reddit/blob/master/r2/r2/lib/db/_sorts.pyx
+	https://steamdb.info/blog/steamdb-rating
+	https://www.evanmiller.org/how-not-to-sort-by-average-rating.html
+	https://redditblog.com/2009/10/15/reddits-new-comment-sorting-system
+	https://www.reddit.com/r/TheoryOfReddit/comments/bpmd3x/how_does_hot_vs_best_vscontroversial_vs_rising/envijlj
+"""
+
+
+# this is the z-score of 0.8, z is calulated via: norm.ppf(1-(1-0.8)/2)
+z_score_08 = norm.ppf(0.9)
+
+
+def _sign(x: Union[int, float]) -> int :
+	return (x > 0) - (x < 0)
+
+
+def hot(up: int, down: int, time: float) -> float :
+	s: int = up - down
+	return _sign(s) * log10(max(abs(s), 1)) + (time - epoch) / 45000
+
+
+def controversial(up: int, down: int) -> float :
+	return (up + down)**(min(up, down)/max(up, down)) if up or down else 0
+
+
+def confidence(up: int, total: int) -> float :
+	# calculates a confidence score with a z score of 0.8
+	if not total :
+		return 0
+	phat = up / total
+	return (
+		(phat + z_score_08 * z_score_08 / (2 * total)
+		- z_score_08 * sqrt((phat * (1 - phat)
+		+ z_score_08 * z_score_08 / (4 * total)) / total)) / (1 + z_score_08 * z_score_08 / total)
+	)
+
+
+def best(up: int, total: int) -> float :
+	if not total :
+		return 0
+	s: float = up / total
+	return s - (s - 0.5) * 2**(-log10(total + 1))
 
 
 ScoreCache: KeyValueStore = KeyValueStore('kheina', 'score')
@@ -60,9 +109,9 @@ class Scoring(SqlInterface) :
 			created: float = data[2].timestamp()
 
 			top: int = up - down
-			hot: float = calc_hot(up, down, created)
+			hot: float = hot(up, down, created)
 			best: float = confidence(up, total)
-			controversial: float = calc_cont(up, down)
+			cont: float = controversial(up, down)
 
 			transaction.query("""
 				INSERT INTO kheina.public.post_scores
@@ -80,23 +129,31 @@ class Scoring(SqlInterface) :
 					WHERE post_scores.post_id = %s;
 				""",
 				(
-					post_id.int(), up, down, top, hot, best, controversial,
-					up, down, top, hot, best, controversial, post_id.int(),
+					post_id.int(), up, down, top, hot, best, cont,
+					up, down, top, hot, best, cont, post_id.int(),
 				),
 			)
 
 			transaction.commit()
 
+		score: Dict[str, int] = {
+			'up': up,
+			'down': down,
+			'total': total,
+		}
+		ScoreCache.put(post_id, score)
+
+		user_vote = 0 if upvote is None else (1 if upvote else -1)
+		VoteCache.put(f'{user.user_id}.{post_id}', user_vote)
+
 		return Score(
-			up = up,
-			down = down,
-			total = total,
-			user_vote = 0 if upvote is None else (1 if upvote else -1),
+			**score,
+			user_vote = user_vote,
 		)
 
 
 	@AerospikeCache('kheina', 'score', '{post_id}', _kvs=ScoreCache)
-	async def _get_score(self, post_id: PostId) -> Dict[str, int] :
+	async def _get_score(self, post_id: PostId) -> Optional[Dict[str, int]] :
 		data: List[int] = await self.query_async("""
 			SELECT
 				post_scores.upvotes,
@@ -109,16 +166,16 @@ class Scoring(SqlInterface) :
 		)
 
 		if not data :
-			raise NotFound(f'no data was found for the provided post id: {post_id}.')
+			return None
 
 		return  {
-			'up':data[0],
+			'up': data[0],
 			'down': data[1],
-			'total': data[0] + data[1],
+			'total': sum(data),
 		}
 
 
-	@AerospikeCache('kheina', 'votes', '{user}.{post_id}', _kvs=ScoreCache)
+	@AerospikeCache('kheina', 'votes', '{user}.{post_id}', _kvs=VoteCache)
 	async def _get_vote(self, user: int, post_id: PostId) -> int :
 		data: List[int] = await self.query_async("""
 			SELECT
@@ -137,11 +194,16 @@ class Scoring(SqlInterface) :
 		return 1 if data[0] else -1
 
 
-	async def getScore(self, user: KhUser, post_id: PostId) -> Score :
-		score: Task[Score] = ensure_future(self._get_score(post_id))
+	async def getScore(self, user: KhUser, post_id: PostId) -> Optional[Score] :
+		score: Task[Dict[str, int]] = ensure_future(self._get_score(post_id))
 		vote: Task[int] = ensure_future(self._get_vote(user.user_id, post_id))
 
+		score: Optional[Dict[str, int]] = await score
+
+		if not score :
+			return None
+
 		return Score(
 			user_vote=await vote,
-			**await score,
+			**score,
 		)
